@@ -229,8 +229,9 @@ int ret;
       gnutls_free(*res);
       return GNUTLS_E_MEMORY_ERROR;
     }
-  (*res)->verify_bits = DEFAULT_VERIFY_BITS;
-  (*res)->verify_depth = DEFAULT_VERIFY_DEPTH;
+  (*res)->verify_bits = DEFAULT_MAX_VERIFY_BITS;
+  (*res)->verify_depth = DEFAULT_MAX_VERIFY_DEPTH;
+  (*res)->verify_flags = GNUTLS_VERIFY_ALLOW_UNSORTED_CHAIN;
 
   return 0;
 }
