@@ -71,14 +71,10 @@ _gnutls_remove_unwanted_ciphersuites (gnutls_session_t session,
                                       int cipher_suites_size,
                                       gnutls_pk_algorithm_t *pk_algos,
                                       size_t pk_algos_size);
+static int _gnutls_handshake_common (gnutls_session_t session);
 static int _gnutls_handshake_client (gnutls_session_t session);
 static int _gnutls_handshake_server (gnutls_session_t session);
 
-static int
-_gnutls_recv_handshake_final (gnutls_session_t session, int init);
-static int
-_gnutls_send_handshake_final (gnutls_session_t session, int init);
-
 /* Empties but does not free the buffer
  */
 static inline void
@@ -777,10 +773,8 @@ server_find_pk_algos_in_ciphersuites (const uint8_t *
                                               gnutls_pk_algorithm_t * algos,
                                               size_t* algos_size)
 {
-  unsigned int j, x;
+  unsigned int j;
   gnutls_kx_algorithm_t kx;
-  gnutls_pk_algorithm_t pk;
-  unsigned found;
   unsigned int max = *algos_size;
 
   if (datalen % 2 != 0)
@@ -795,23 +789,10 @@ server_find_pk_algos_in_ciphersuites (const uint8_t *
       kx = _gnutls_cipher_suite_get_kx_algo (&data[j]);
       if (_gnutls_map_kx_get_cred (kx, 1) == GNUTLS_CRD_CERTIFICATE)
         {
-          pk = _gnutls_map_pk_get_pk (kx);
-          found = 0;
-          for (x=0;x<*algos_size;x++)
-            {
-              if (algos[x] == pk) 
-                {
-                  found = 1;
-                  break;
-                }
-            }
+          algos[(*algos_size)++] = _gnutls_map_pk_get_pk (kx);
           
-          if (found == 0)
-            {
-              algos[(*algos_size)++] = _gnutls_map_pk_get_pk (kx);
-              if ((*algos_size) >= max)
-                return 0;
-            }
+          if ((*algos_size) >= max)
+            return 0;
         }
     }
 
@@ -867,7 +848,7 @@ _gnutls_server_select_suite (gnutls_session_t session, uint8_t * data,
   ret = _gnutls_supported_ciphersuites (session, cipher_suites, sizeof(cipher_suites));
   if (ret < 0)
     return gnutls_assert_val(ret);
-
+    
   cipher_suites_size = ret;
 
   /* Here we remove any ciphersuite that does not conform
@@ -1172,8 +1153,8 @@ _gnutls_send_handshake (gnutls_session_t session, mbuffer_st * bufel,
   switch (type)
     {
     case GNUTLS_HANDSHAKE_CERTIFICATE_PKT:     /* this one is followed by ServerHelloDone
-                                                * or ClientKeyExchange always.
-                                                */
+                                                 * or ClientKeyExchange always.
+                                                 */
     case GNUTLS_HANDSHAKE_SERVER_KEY_EXCHANGE: /* as above */
     case GNUTLS_HANDSHAKE_SERVER_HELLO:        /* as above */
     case GNUTLS_HANDSHAKE_CERTIFICATE_REQUEST: /* as above */
@@ -1259,16 +1240,13 @@ _gnutls_handshake_hash_add_sent (gnutls_session_t session,
       if (gnutls_protocol_get_version (session) == GNUTLS_DTLS0_9) 
         {
 	  /* Old DTLS doesn't include the header in the MAC */
-	  if (datalen < 12) 
+	  if (datalen <= 12) 
 	    {
 	      gnutls_assert ();
-	      return GNUTLS_E_INTERNAL_ERROR;
+	      return GNUTLS_E_INVALID_REQUEST;
 	    }
 	  dataptr += 12;
 	  datalen -= 12;
-	  
-	  if (datalen == 0)
-	  	return 0;
         }
 
       ret = _gnutls_buffer_append_data(&session->internals.handshake_hash_buffer,
@@ -1309,8 +1287,6 @@ _gnutls_recv_handshake (gnutls_session_t session,
       return gnutls_assert_val_fatal(ret);
     }
 
-  session->internals.last_handshake_in = hsk.htype;
-
   ret = _gnutls_handshake_hash_add_recvd (session, hsk.htype,
                                               hsk.header, hsk.header_size,
                                               hsk.data.data, hsk.data.length);
@@ -2391,6 +2367,11 @@ gnutls_handshake (gnutls_session_t session)
   if (session->internals.priorities.protocol.algorithms == 0)
     return gnutls_assert_val(GNUTLS_E_NO_PRIORITIES_WERE_SET);
 
+  if (session->internals.handshake_timeout_ms && 
+      session->internals.handshake_endtime == 0)
+    session->internals.handshake_endtime = gnutls_time(0) + 
+      session->internals.handshake_timeout_ms / 1000;
+
   ret = _gnutls_epoch_get (session, session->security_parameters.epoch_next,
                            &params);
   if (ret < 0)
@@ -2425,8 +2406,17 @@ gnutls_handshake (gnutls_session_t session)
       return ret;
     }
 
-  /* clear handshake buffer */
-  _gnutls_handshake_hash_buffers_clear (session);
+  ret = _gnutls_handshake_common (session);
+
+  if (ret < 0)
+    {
+      if (_gnutls_abort_handshake (session, ret) == 0)
+        STATE = STATE0;
+
+      return ret;
+    }
+
+  STATE = STATE0;
 
   if (IS_DTLS(session)==0)
     {
@@ -2444,6 +2434,28 @@ gnutls_handshake (gnutls_session_t session)
   return 0;
 }
 
+/**
+ * gnutls_handshake_set_timeout:
+ * @session: is a #gnutls_session_t structure.
+ * @ms: is a timeout value in milliseconds
+ *
+ * This function sets the timeout for the handshake process
+ * to the provided value. Use an @ms value of zero to disable
+ * timeout.
+ *
+ * Note that in order for the timeout to be enforced
+ * gnutls_transport_set_pull_timeout_function() must be set.
+ *
+ **/
+void
+gnutls_handshake_set_timeout (gnutls_session_t session, unsigned int ms)
+{
+  if (ms == GNUTLS_DEFAULT_HANDSHAKE_TIMEOUT)
+    ms = 40*1000;
+  session->internals.handshake_timeout_ms = ms;
+}
+
+
 #define IMED_RET( str, ret, allow_alert) do { \
 	if (ret < 0) { \
 		/* EAGAIN and INTERRUPTED are always non-fatal */ \
@@ -2458,6 +2470,7 @@ gnutls_handshake (gnutls_session_t session)
 	} } while (0)
 
 
+
 /*
  * _gnutls_handshake_client 
  * This function performs the client side of the handshake of the TLS/SSL protocol.
@@ -2487,14 +2500,14 @@ _gnutls_handshake_client (gnutls_session_t session)
       STATE = STATE1;
       IMED_RET ("send hello", ret, 1);
 
-    case STATE2:
+    case STATE11:
       if (IS_DTLS (session))
         {
           ret =
             _gnutls_recv_handshake (session, 
                   GNUTLS_HANDSHAKE_HELLO_VERIFY_REQUEST,
                   1, NULL);
-          STATE = STATE2;
+          STATE = STATE11;
           IMED_RET ("recv hello verify", ret, 1);
 
           if (ret == 1)
@@ -2503,130 +2516,91 @@ _gnutls_handshake_client (gnutls_session_t session)
               return 1;
             }
         }
-    case STATE3:
+    case STATE2:
       /* receive the server hello */
       ret =
         _gnutls_recv_handshake (session,
                                 GNUTLS_HANDSHAKE_SERVER_HELLO,
                                 0, NULL);
-      STATE = STATE3;
+      STATE = STATE2;
       IMED_RET ("recv hello", ret, 1);
 
-    case STATE4:
+    case STATE70:
       if (session->security_parameters.do_recv_supplemental)
         {
           ret = _gnutls_recv_supplemental (session);
-          STATE = STATE4;
+          STATE = STATE70;
           IMED_RET ("recv supplemental", ret, 1);
         }
 
-    case STATE5:
+    case STATE3:
       /* RECV CERTIFICATE */
       if (session->internals.resumed == RESUME_FALSE)   /* if we are not resuming */
         ret = _gnutls_recv_server_certificate (session);
-      STATE = STATE5;
+      STATE = STATE3;
       IMED_RET ("recv server certificate", ret, 1);
 
-    case STATE8:
+    case STATE4:
       /* receive the server key exchange */
       if (session->internals.resumed == RESUME_FALSE)   /* if we are not resuming */
         ret = _gnutls_recv_server_kx_message (session);
-      STATE = STATE8;
+      STATE = STATE4;
       IMED_RET ("recv server kx message", ret, 1);
 
-    case STATE9:
+    case STATE5:
       /* receive the server certificate request - if any 
        */
 
       if (session->internals.resumed == RESUME_FALSE)   /* if we are not resuming */
         ret = _gnutls_recv_server_crt_request (session);
-      STATE = STATE9;
+      STATE = STATE5;
       IMED_RET ("recv server certificate request message", ret, 1);
 
-    case STATE10:
+    case STATE6:
       /* receive the server hello done */
       if (session->internals.resumed == RESUME_FALSE)   /* if we are not resuming */
         ret =
           _gnutls_recv_handshake (session,
                                   GNUTLS_HANDSHAKE_SERVER_HELLO_DONE,
                                   0, NULL);
-      STATE = STATE10;
+      STATE = STATE6;
       IMED_RET ("recv server hello done", ret, 1);
-
-    case STATE11:
+    case STATE71:
       if (session->security_parameters.do_send_supplemental)
         {
-          ret = _gnutls_send_supplemental (session, AGAIN (STATE11));
-          STATE = STATE11;
+          ret = _gnutls_send_supplemental (session, AGAIN (STATE71));
+          STATE = STATE71;
           IMED_RET ("send supplemental", ret, 0);
         }
 
-    case STATE12:
+    case STATE7:
       /* send our certificate - if any and if requested
        */
       if (session->internals.resumed == RESUME_FALSE)   /* if we are not resuming */
-        ret = _gnutls_send_client_certificate (session, AGAIN (STATE12));
-      STATE = STATE12;
+        ret = _gnutls_send_client_certificate (session, AGAIN (STATE7));
+      STATE = STATE7;
       IMED_RET ("send client certificate", ret, 0);
 
-    case STATE13:
+    case STATE8:
       if (session->internals.resumed == RESUME_FALSE)   /* if we are not resuming */
-        ret = _gnutls_send_client_kx_message (session, AGAIN (STATE13));
-      STATE = STATE13;
+        ret = _gnutls_send_client_kx_message (session, AGAIN (STATE8));
+      STATE = STATE8;
       IMED_RET ("send client kx", ret, 0);
 
-    case STATE14:
+    case STATE9:
       /* send client certificate verify */
       if (session->internals.resumed == RESUME_FALSE)   /* if we are not resuming */
         ret =
-          _gnutls_send_client_certificate_verify (session, AGAIN (STATE14));
-      STATE = STATE14;
+          _gnutls_send_client_certificate_verify (session, AGAIN (STATE9));
+      STATE = STATE9;
       IMED_RET ("send client certificate verify", ret, 1);
 
-    case STATE15:
-      STATE = STATE15;
-      if (session->internals.resumed == RESUME_FALSE)
-        {
-          ret = _gnutls_send_handshake_final (session, TRUE);
-          IMED_RET ("send handshake final 2", ret, 1);
-        }
-      else
-        {
-          ret = _gnutls_recv_new_session_ticket (session);
-          IMED_RET ("recv handshake new session ticket", ret, 1);
-        }
-
-    case STATE16:
-      STATE = STATE16;
-      if (session->internals.resumed == RESUME_FALSE)
-        {
-          ret = _gnutls_recv_new_session_ticket (session);
-          IMED_RET ("recv handshake new session ticket", ret, 1);
-        }
-      else
-        {
-          ret = _gnutls_recv_handshake_final (session, TRUE);
-          IMED_RET ("recv handshake final", ret, 1);
-        }
-
-    case STATE17:
-      STATE = STATE17;
-      if (session->internals.resumed == RESUME_FALSE)
-        {
-          ret = _gnutls_recv_handshake_final (session, FALSE);
-          IMED_RET ("recv handshake final 2", ret, 1);
-        }
-      else
-        {
-          ret = _gnutls_send_handshake_final (session, FALSE);
-          IMED_RET ("send handshake final", ret, 1);
-        }
-
       STATE = STATE0;
     default:
       break;
     }
 
+
   return 0;
 }
 
@@ -2685,12 +2659,12 @@ _gnutls_send_handshake_final (gnutls_session_t session, int init)
 
   /* Send the CHANGE CIPHER SPEC PACKET */
 
-  switch (FINAL_STATE)
+  switch (STATE)
     {
     case STATE0:
-    case STATE1:
-      ret = send_change_cipher_spec (session, FAGAIN (STATE1));
-      FINAL_STATE = STATE0;
+    case STATE20:
+      ret = send_change_cipher_spec (session, AGAIN (STATE20));
+      STATE = STATE0;
 
       if (ret < 0)
         {
@@ -2717,10 +2691,10 @@ _gnutls_send_handshake_final (gnutls_session_t session, int init)
           return ret;
         }
 
-    case STATE2:
+    case STATE21:
       /* send the finished message */
-      ret = _gnutls_send_finished (session, FAGAIN (STATE2));
-      FINAL_STATE = STATE2;
+      ret = _gnutls_send_finished (session, AGAIN (STATE21));
+      STATE = STATE21;
       if (ret < 0)
         {
           ERR ("send Finished", ret);
@@ -2728,7 +2702,7 @@ _gnutls_send_handshake_final (gnutls_session_t session, int init)
           return ret;
         }
 
-      FINAL_STATE = STATE0;
+      STATE = STATE0;
     default:
       break;
     }
@@ -2746,12 +2720,18 @@ _gnutls_recv_handshake_final (gnutls_session_t session, int init)
   int ret = 0;
   uint8_t ch;
   unsigned int ccs_len = 1;
+  unsigned int tleft;
   
-  switch (FINAL_STATE)
+  ret = handshake_remaining_time(session);
+  if (ret < 0)
+    return gnutls_assert_val(ret);
+  tleft = ret;
+
+  switch (STATE)
     {
     case STATE0:
     case STATE30:
-      FINAL_STATE = STATE30;
+      STATE = STATE30;
 
       /* This is the last flight and peer cannot be sure
        * we have received it unless we notify him. So we
@@ -2768,7 +2748,8 @@ _gnutls_recv_handshake_final (gnutls_session_t session, int init)
       if (gnutls_protocol_get_version (session) == GNUTLS_DTLS0_9)
         ccs_len = 3;
 
-      ret = _gnutls_recv_int (session, GNUTLS_CHANGE_CIPHER_SPEC, -1, &ch, ccs_len, NULL);
+      ret = _gnutls_recv_int (session, GNUTLS_CHANGE_CIPHER_SPEC, -1, &ch, ccs_len, NULL, 
+                              tleft);
       if (ret <= 0)
         {
           ERR ("recv ChangeCipherSpec", ret);
@@ -2798,7 +2779,7 @@ _gnutls_recv_handshake_final (gnutls_session_t session, int init)
         }
 
     case STATE31:
-      FINAL_STATE = STATE31;
+      STATE = STATE31;
 
        if (IS_DTLS(session) && !_dtls_is_async(session) && 
            (gnutls_record_check_pending( session) +
@@ -2816,7 +2797,7 @@ _gnutls_recv_handshake_final (gnutls_session_t session, int init)
           gnutls_assert ();
           return ret;
         }
-      FINAL_STATE = STATE0;
+      STATE = STATE0;
     default:
       break;
     }
@@ -2867,29 +2848,29 @@ _gnutls_handshake_server (gnutls_session_t session)
       STATE = STATE3;
       IMED_RET ("send server certificate", ret, 0);
 
-    case STATE5:
+    case STATE4:
       /* send server key exchange (A) */
       if (session->internals.resumed == RESUME_FALSE)
-        ret = _gnutls_send_server_kx_message (session, AGAIN (STATE5));
-      STATE = STATE5;
+        ret = _gnutls_send_server_kx_message (session, AGAIN (STATE4));
+      STATE = STATE4;
       IMED_RET ("send server kx", ret, 0);
 
-    case STATE6:
+    case STATE5:
       /* Send certificate request - if requested to */
       if (session->internals.resumed == RESUME_FALSE)
         ret =
-          _gnutls_send_server_crt_request (session, AGAIN (STATE6));
-      STATE = STATE6;
+          _gnutls_send_server_crt_request (session, AGAIN (STATE5));
+      STATE = STATE5;
       IMED_RET ("send server cert request", ret, 0);
 
-    case STATE7:
+    case STATE6:
       /* send the server hello done */
       if (session->internals.resumed == RESUME_FALSE)   /* if we are not resuming */
         ret =
           _gnutls_send_empty_handshake (session,
                                         GNUTLS_HANDSHAKE_SERVER_HELLO_DONE,
-                                        AGAIN (STATE7));
-      STATE = STATE7;
+                                        AGAIN (STATE6));
+      STATE = STATE6;
       IMED_RET ("send server hello done", ret, 1);
 
     case STATE71:
@@ -2901,65 +2882,28 @@ _gnutls_handshake_server (gnutls_session_t session)
         }
 
       /* RECV CERTIFICATE + KEYEXCHANGE + CERTIFICATE_VERIFY */
-    case STATE8:
+    case STATE7:
       /* receive the client certificate message */
       if (session->internals.resumed == RESUME_FALSE)   /* if we are not resuming */
         ret = _gnutls_recv_client_certificate (session);
-      STATE = STATE8;
+      STATE = STATE7;
       IMED_RET ("recv client certificate", ret, 1);
 
-    case STATE10:
+    case STATE8:
       /* receive the client key exchange message */
       if (session->internals.resumed == RESUME_FALSE)   /* if we are not resuming */
         ret = _gnutls_recv_client_kx_message (session);
-      STATE = STATE10;
+      STATE = STATE8;
       IMED_RET ("recv client kx", ret, 1);
 
-    case STATE11:
+    case STATE9:
       /* receive the client certificate verify message */
       if (session->internals.resumed == RESUME_FALSE)   /* if we are not resuming */
         ret = _gnutls_recv_client_certificate_verify_message (session);
-      STATE = STATE11;
+      STATE = STATE9;
       IMED_RET ("recv client certificate verify", ret, 1);
 
-    case STATE12:
-      STATE = STATE12;
-      if (session->internals.resumed == RESUME_FALSE)   /* if we are not resuming */
-        {
-          ret = _gnutls_recv_handshake_final (session, TRUE);
-          IMED_RET ("recv handshake final", ret, 1);
-        }
-      else
-        {
-          ret = _gnutls_send_handshake_final (session, TRUE);
-          IMED_RET ("send handshake final 2", ret, 1);
-        }
-
-    case STATE13:
-        ret = _gnutls_send_new_session_ticket (session, AGAIN (STATE13));
-        STATE = STATE13;
-        IMED_RET ("send handshake new session ticket", ret, 0);
-
-    case STATE14:
-      STATE = STATE14;
-      if (session->internals.resumed == RESUME_FALSE)   /* if we are not resuming */
-        {
-          ret = _gnutls_send_handshake_final (session, FALSE);
-          IMED_RET ("send handshake final", ret, 1);
-        }
-      else
-        {
-          ret = _gnutls_recv_handshake_final (session, FALSE);
-          IMED_RET ("recv handshake final 2", ret, 1);
-        }
-
-      if (session->security_parameters.entity == GNUTLS_SERVER && session->internals.ticket_sent == 0)
-        {
-          /* if no ticket, save session data */
-          _gnutls_server_register_current_session (session);
-        }
-
-      STATE = STATE0;
+      STATE = STATE0;           /* finished thus clear session */
     default:
       break;
     }
@@ -2967,6 +2911,73 @@ _gnutls_handshake_server (gnutls_session_t session)
   return 0;
 }
 
+static int
+_gnutls_handshake_common (gnutls_session_t session)
+{
+  int ret = 0;
+
+  /* send and recv the change cipher spec and finished messages */
+  if ((session->internals.resumed != RESUME_FALSE
+       && session->security_parameters.entity == GNUTLS_CLIENT)
+      || (session->internals.resumed == RESUME_FALSE
+          && session->security_parameters.entity == GNUTLS_SERVER))
+    {
+      /* if we are a client resuming - or we are a server not resuming */
+      ret = _gnutls_recv_handshake_final (session, TRUE);
+      IMED_RET ("recv handshake final", ret, 1);
+
+      switch (STATE)
+        {
+        case STATE0:
+        case STATE40:
+          ret = _gnutls_send_new_session_ticket (session, AGAIN (STATE40));
+          STATE = STATE40;
+          IMED_RET ("send handshake new session ticket", ret, 0);
+          STATE = STATE0;
+        default:
+          break;
+        }
+
+      ret = _gnutls_send_handshake_final (session, FALSE);
+      IMED_RET ("send handshake final", ret, 1);
+
+      /* only store if we are not resuming a session and we didn't previously send a ticket 
+       */
+      if (session->security_parameters.entity == GNUTLS_SERVER && session->internals.ticket_sent == 0)
+        {
+          /* in order to support session resuming */
+          _gnutls_server_register_current_session (session);
+        }
+    }
+  else
+    {                           /* if we are a client not resuming - or we are a server resuming */
+
+      ret = _gnutls_send_handshake_final (session, TRUE);
+      IMED_RET ("send handshake final 2", ret, 1);
+
+      switch (STATE)
+        {
+        case STATE0:
+        case STATE41:
+          ret = _gnutls_recv_new_session_ticket (session);
+          STATE = STATE41;
+          IMED_RET ("recv handshake new session ticket", ret, 1);
+          STATE = STATE0;
+        default:
+          break;
+        }
+
+      ret = _gnutls_recv_handshake_final (session, FALSE);
+      IMED_RET ("recv handshake final 2", ret, 1);
+
+    }
+
+
+  /* clear handshake buffer */
+  _gnutls_handshake_hash_buffers_clear (session);
+  return ret;
+
+}
 
 int
 _gnutls_generate_session_id (uint8_t * session_id, uint8_t * len)
@@ -3155,6 +3166,7 @@ _gnutls_remove_unwanted_ciphersuites (gnutls_session_t session,
    * by that certificate and are on the same authentication
    * method (CERTIFICATE).
    */
+
   cert_cred =
     (gnutls_certificate_credentials_t) _gnutls_get_cred (session->key,
                                                          GNUTLS_CRD_CERTIFICATE,
