@@ -45,10 +45,6 @@ sign_tls_hash (gnutls_session_t session, gnutls_digest_algorithm_t hash_algo,
                   const gnutls_datum_t * hash_concat,
                   gnutls_datum_t * signature);
 
-static int
-encode_ber_digest_info (gnutls_digest_algorithm_t hash,
-                        const gnutls_datum_t * digest,
-                        gnutls_datum_t * output);
 
 /* While this is currently equal to the length of RSA/SHA512
  * signature, it should also be sufficient for DSS signature and any
@@ -80,8 +76,10 @@ _gnutls_handshake_sign_data (gnutls_session_t session, gnutls_pcert_st* cert,
       gnutls_assert ();
       return GNUTLS_E_UNKNOWN_PK_ALGORITHM;
     }
+    
+  gnutls_sign_algorithm_set(session, *sign_algo);
 
-  hash_algo = _gnutls_sign_get_hash_algorithm (*sign_algo);
+  hash_algo = gnutls_sign_get_hash_algorithm (*sign_algo);
 
   _gnutls_handshake_log ("HSK[%p]: signing handshake data: using %s\n",
                     session, gnutls_sign_algorithm_get_name (*sign_algo));
@@ -99,7 +97,7 @@ _gnutls_handshake_sign_data (gnutls_session_t session, gnutls_pcert_st* cert,
                 GNUTLS_RANDOM_SIZE);
   _gnutls_hash (&td_sha, params->data, params->size);
 
-  switch (gnutls_pubkey_get_pk_algorithm(cert->pubkey, NULL))
+  switch (gnutls_privkey_get_pk_algorithm(pkey, NULL))
     {
     case GNUTLS_PK_RSA:
       if (!_gnutls_version_has_selectable_sighash (ver))
@@ -163,41 +161,6 @@ _gnutls_handshake_sign_data (gnutls_session_t session, gnutls_pcert_st* cert,
 
 }
 
-
-/* This will create a PKCS1 or DSA signature, using the given parameters, and the
- * given data. The output will be allocated and be put in signature.
- */
-int
-_gnutls_soft_sign (gnutls_pk_algorithm_t algo, gnutls_pk_params_st * params,
-                   const gnutls_datum_t * data,
-                   gnutls_datum_t * signature)
-{
-  int ret;
-
-  switch (algo)
-    {
-    case GNUTLS_PK_RSA:
-      /* encrypt */
-      if ((ret = _gnutls_pkcs1_rsa_encrypt (signature, data, params, 1)) < 0)
-        {
-          gnutls_assert ();
-          return ret;
-        }
-
-      break;
-    default:
-      ret = _gnutls_pk_sign( algo, signature, data, params);
-      if (ret < 0)
-        {
-          gnutls_assert ();
-          return ret;
-        }
-      break;
-    }
-
-  return 0;
-}
-
 /* This will create a PKCS1 or DSA signature, as defined in the TLS protocol.
  * Cert is the certificate of the corresponding private key. It is only checked if
  * it supports signing.
@@ -221,7 +184,10 @@ sign_tls_hash (gnutls_session_t session, gnutls_digest_algorithm_t hash_algo,
         if (!(key_usage & GNUTLS_KEY_DIGITAL_SIGNATURE))
           {
             gnutls_assert ();
-            _gnutls_audit_log(session, "Peer's certificate does not allow digital signatures. Key usage violation detected (ignored).\n");
+            if (session->internals.priorities.allow_key_usage_violation == 0)
+              return GNUTLS_E_KEY_USAGE_VIOLATION;
+            else
+              _gnutls_audit_log(session, "Key usage violation was detected (ignored).\n");
           }
 
       /* External signing. Deprecated. To be removed. */
@@ -269,11 +235,12 @@ es_cleanup:
 }
 
 static int
-verify_tls_hash (gnutls_session_t session, gnutls_protocol_t ver, gnutls_pcert_st* cert,
-                    const gnutls_datum_t * hash_concat,
-                    gnutls_datum_t * signature, size_t sha1pos,
-                    gnutls_sign_algorithm_t sign_algo,
-                    gnutls_pk_algorithm_t pk_algo)
+verify_tls_hash (gnutls_session_t session,
+                 gnutls_protocol_t ver, gnutls_pcert_st* cert,
+                 const gnutls_datum_t * hash_concat,
+                 gnutls_datum_t * signature, size_t sha1pos,
+                 gnutls_sign_algorithm_t sign_algo,
+                 gnutls_pk_algorithm_t pk_algo)
 {
   int ret;
   gnutls_datum_t vdata;
@@ -293,7 +260,10 @@ verify_tls_hash (gnutls_session_t session, gnutls_protocol_t ver, gnutls_pcert_s
     if (!(key_usage & GNUTLS_KEY_DIGITAL_SIGNATURE))
       {
         gnutls_assert ();
-        _gnutls_audit_log(session, "Peer's certificate does not allow digital signatures. Key usage violation detected (ignored).\n");
+        if (session->internals.priorities.allow_key_usage_violation == 0)
+          return GNUTLS_E_KEY_USAGE_VIOLATION;
+        else
+          _gnutls_audit_log(session, "Key usage violation was detected (ignored).\n");
       }
 
   if (pk_algo == GNUTLS_PK_UNKNOWN)
@@ -310,12 +280,9 @@ verify_tls_hash (gnutls_session_t session, gnutls_protocol_t ver, gnutls_pcert_s
         flags = GNUTLS_PUBKEY_VERIFY_FLAG_TLS_RSA;
       else
         flags = 0;
-
-
       break;
     case GNUTLS_PK_DSA:
     case GNUTLS_PK_EC:
-
       vdata.data = &hash_concat->data[sha1pos];
       vdata.size = hash_concat->size - sha1pos;
 
@@ -368,7 +335,7 @@ _gnutls_handshake_verify_data (gnutls_session_t session, gnutls_pcert_st* cert,
       if (ret < 0)
         return gnutls_assert_val(ret);
 
-      hash_algo = _gnutls_sign_get_hash_algorithm (sign_algo);
+      hash_algo = gnutls_sign_get_hash_algorithm (sign_algo);
     }
   else
     {
@@ -419,10 +386,8 @@ _gnutls_handshake_verify_data (gnutls_session_t session, gnutls_pcert_st* cert,
     }
 
   ret = verify_tls_hash (session, ver, cert, &dconcat, signature,
-                            dconcat.size -
-                            _gnutls_hash_get_algo_len (hash_algo),
-                            sign_algo,
-                            _gnutls_sign_get_pk_algorithm (sign_algo));
+                            dconcat.size - _gnutls_hash_get_algo_len (hash_algo),
+                            sign_algo, gnutls_sign_get_pk_algorithm (sign_algo));
   if (ret < 0)
     {
       gnutls_assert ();
@@ -455,7 +420,7 @@ _gnutls_handshake_verify_crt_vrfy12 (gnutls_session_t session,
   if (ret < 0)
     return gnutls_assert_val(ret);
   
-  hash_algo = _gnutls_sign_get_hash_algorithm(sign_algo);
+  hash_algo = gnutls_sign_get_hash_algorithm(sign_algo);
   
   ret = _gnutls_hash_fast(hash_algo, session->internals.handshake_hash_buffer.data, 
                           session->internals.handshake_hash_buffer_prev_len,
@@ -500,7 +465,7 @@ _gnutls_handshake_verify_crt_vrfy (gnutls_session_t session,
 
   if (_gnutls_version_has_selectable_sighash(ver))
     return _gnutls_handshake_verify_crt_vrfy12 (session, cert, signature,
-                                                 sign_algo);
+                                                sign_algo);
 
   ret =
     _gnutls_hash_init (&td_md5, GNUTLS_DIG_MD5);
@@ -562,8 +527,8 @@ _gnutls_handshake_verify_crt_vrfy (gnutls_session_t session,
 
   ret =
     verify_tls_hash (session, ver, cert, &dconcat, signature, 16,
-                        GNUTLS_SIGN_UNKNOWN,
-                        gnutls_pubkey_get_pk_algorithm(cert->pubkey, NULL));
+                     GNUTLS_SIGN_UNKNOWN,
+                     gnutls_pubkey_get_pk_algorithm(cert->pubkey, NULL));
   if (ret < 0)
     {
       gnutls_assert ();
@@ -594,8 +559,10 @@ _gnutls_handshake_sign_crt_vrfy12 (gnutls_session_t session,
       gnutls_assert ();
       return GNUTLS_E_UNKNOWN_PK_ALGORITHM;
     }
+  
+  gnutls_sign_algorithm_set(session, sign_algo);
 
-  hash_algo = _gnutls_sign_get_hash_algorithm (sign_algo);
+  hash_algo = gnutls_sign_get_hash_algorithm (sign_algo);
 
   _gnutls_debug_log ("sign handshake cert vrfy: picked %s with %s\n",
                     gnutls_sign_algorithm_get_name (sign_algo),
@@ -640,11 +607,11 @@ _gnutls_handshake_sign_crt_vrfy (gnutls_session_t session,
   digest_hd_st td_md5;
   digest_hd_st td_sha;
   gnutls_protocol_t ver = gnutls_protocol_get_version (session);
-  gnutls_pk_algorithm_t pk = gnutls_pubkey_get_pk_algorithm(cert->pubkey, NULL);
+  gnutls_pk_algorithm_t pk = gnutls_privkey_get_pk_algorithm(pkey, NULL);
 
   if (_gnutls_version_has_selectable_sighash(ver))
     return _gnutls_handshake_sign_crt_vrfy12 (session, cert, pkey,
-                                                 signature);
+                                              signature);
 
   ret =
     _gnutls_hash_init (&td_sha, GNUTLS_DIG_SHA1);
@@ -791,170 +758,3 @@ pk_prepare_hash (gnutls_pk_algorithm_t pk,
   return 0;
 }
 
-/* Reads the digest information.
- * we use DER here, although we should use BER. It works fine
- * anyway.
- */
-int
-decode_ber_digest_info (const gnutls_datum_t * info,
-                        gnutls_digest_algorithm_t * hash,
-                        uint8_t * digest, unsigned int *digest_size)
-{
-  ASN1_TYPE dinfo = ASN1_TYPE_EMPTY;
-  int result;
-  char str[1024];
-  int len;
-
-  if ((result = asn1_create_element (_gnutls_get_gnutls_asn (),
-                                     "GNUTLS.DigestInfo",
-                                     &dinfo)) != ASN1_SUCCESS)
-    {
-      gnutls_assert ();
-      return _gnutls_asn2err (result);
-    }
-
-  result = asn1_der_decoding (&dinfo, info->data, info->size, NULL);
-  if (result != ASN1_SUCCESS)
-    {
-      gnutls_assert ();
-      asn1_delete_structure (&dinfo);
-      return _gnutls_asn2err (result);
-    }
-
-  len = sizeof (str) - 1;
-  result = asn1_read_value (dinfo, "digestAlgorithm.algorithm", str, &len);
-  if (result != ASN1_SUCCESS)
-    {
-      gnutls_assert ();
-      asn1_delete_structure (&dinfo);
-      return _gnutls_asn2err (result);
-    }
-
-  *hash = _gnutls_x509_oid_to_digest (str);
-
-  if (*hash == GNUTLS_DIG_UNKNOWN)
-    {
-
-      _gnutls_debug_log ("verify.c: HASH OID: %s\n", str);
-
-      gnutls_assert ();
-      asn1_delete_structure (&dinfo);
-      return GNUTLS_E_UNKNOWN_ALGORITHM;
-    }
-
-  len = sizeof (str) - 1;
-  result = asn1_read_value (dinfo, "digestAlgorithm.parameters", str, &len);
-  /* To avoid permitting garbage in the parameters field, either the
-     parameters field is not present, or it contains 0x05 0x00. */
-  if (!(result == ASN1_ELEMENT_NOT_FOUND ||
-        (result == ASN1_SUCCESS && len == ASN1_NULL_SIZE &&
-         memcmp (str, ASN1_NULL, ASN1_NULL_SIZE) == 0)))
-    {
-      gnutls_assert ();
-      asn1_delete_structure (&dinfo);
-      return GNUTLS_E_ASN1_GENERIC_ERROR;
-    }
-
-  len = *digest_size;
-  result = asn1_read_value (dinfo, "digest", digest, &len);
-  
-  if (result != ASN1_SUCCESS)
-    {
-      gnutls_assert ();
-      *digest_size = len;
-      asn1_delete_structure (&dinfo);
-      return _gnutls_asn2err (result);
-    }
-
-  *digest_size = len;
-  asn1_delete_structure (&dinfo);
-
-  return 0;
-}
-
-/* Writes the digest information and the digest in a DER encoded
- * structure. The digest info is allocated and stored into the info structure.
- */
-static int
-encode_ber_digest_info (gnutls_digest_algorithm_t hash,
-                        const gnutls_datum_t * digest,
-                        gnutls_datum_t * output)
-{
-  ASN1_TYPE dinfo = ASN1_TYPE_EMPTY;
-  int result;
-  const char *algo;
-  uint8_t *tmp_output;
-  int tmp_output_size;
-
-  algo = _gnutls_x509_mac_to_oid ((gnutls_mac_algorithm_t) hash);
-  if (algo == NULL)
-    {
-      gnutls_assert ();
-      _gnutls_debug_log ("Hash algorithm: %d has no OID\n", hash);
-      return GNUTLS_E_UNKNOWN_PK_ALGORITHM;
-    }
-
-  if ((result = asn1_create_element (_gnutls_get_gnutls_asn (),
-                                     "GNUTLS.DigestInfo",
-                                     &dinfo)) != ASN1_SUCCESS)
-    {
-      gnutls_assert ();
-      return _gnutls_asn2err (result);
-    }
-
-  result = asn1_write_value (dinfo, "digestAlgorithm.algorithm", algo, 1);
-  if (result != ASN1_SUCCESS)
-    {
-      gnutls_assert ();
-      asn1_delete_structure (&dinfo);
-      return _gnutls_asn2err (result);
-    }
-
-  /* Write an ASN.1 NULL in the parameters field.  This matches RFC
-     3279 and RFC 4055, although is arguable incorrect from a historic
-     perspective (see those documents for more information).
-     Regardless of what is correct, this appears to be what most
-     implementations do.  */
-  result = asn1_write_value (dinfo, "digestAlgorithm.parameters",
-                             ASN1_NULL, ASN1_NULL_SIZE);
-  if (result != ASN1_SUCCESS)
-    {
-      gnutls_assert ();
-      asn1_delete_structure (&dinfo);
-      return _gnutls_asn2err (result);
-    }
-
-  result = asn1_write_value (dinfo, "digest", digest->data, digest->size);
-  if (result != ASN1_SUCCESS)
-    {
-      gnutls_assert ();
-      asn1_delete_structure (&dinfo);
-      return _gnutls_asn2err (result);
-    }
-
-  tmp_output_size = 0;
-  asn1_der_coding (dinfo, "", NULL, &tmp_output_size, NULL);
-
-  tmp_output = gnutls_malloc (tmp_output_size);
-  if (tmp_output == NULL)
-    {
-      gnutls_assert ();
-      asn1_delete_structure (&dinfo);
-      return GNUTLS_E_MEMORY_ERROR;
-    }
-
-  result = asn1_der_coding (dinfo, "", tmp_output, &tmp_output_size, NULL);
-  if (result != ASN1_SUCCESS)
-    {
-      gnutls_assert ();
-      asn1_delete_structure (&dinfo);
-      return _gnutls_asn2err (result);
-    }
-
-  asn1_delete_structure (&dinfo);
-
-  output->size = tmp_output_size;
-  output->data = tmp_output;
-
-  return 0;
-}
