@@ -293,7 +293,7 @@ _gnutls_handshake_internal_state_clear (gnutls_session_t session)
  * @flags can be one of %GNUTLS_CLIENT and %GNUTLS_SERVER. For a DTLS
  * entity, the flags %GNUTLS_DATAGRAM and  %GNUTLS_NONBLOCK are
  * also available. The latter flag will enable a non-blocking
- * operation of the DTLS timers.
+ * operation of the DTLS timers. 
  *
  * Returns: %GNUTLS_E_SUCCESS on success, or an error code.
  **/
@@ -326,6 +326,8 @@ gnutls_init (gnutls_session_t * session, unsigned int flags)
 
   /* Initialize buffers */
   _gnutls_buffer_init (&(*session)->internals.handshake_hash_buffer);
+  _gnutls_buffer_init (&(*session)->internals.hb_remote_data);
+  _gnutls_buffer_init (&(*session)->internals.hb_local_data);
 
   _mbuffer_head_init (&(*session)->internals.record_buffer);
   _mbuffer_head_init (&(*session)->internals.record_send_buffer);
@@ -381,6 +383,9 @@ gnutls_init (gnutls_session_t * session, unsigned int flags)
   gnutls_transport_set_errno_function (*session, system_errno);
   gnutls_transport_set_pull_timeout_function (*session, system_recv_timeout);
 
+  (*session)->internals.hb_retrans_timeout_ms = 1000;
+  (*session)->internals.hb_total_timeout_ms = 60000;
+
   if (flags & GNUTLS_DATAGRAM)
     {
       (*session)->internals.dtls.mtu = DTLS_DEFAULT_MTU;
@@ -440,6 +445,9 @@ gnutls_deinit (gnutls_session_t session)
       }
 
   _gnutls_buffer_clear (&session->internals.handshake_hash_buffer);
+  _gnutls_buffer_clear (&session->internals.hb_remote_data);
+  _gnutls_buffer_clear (&session->internals.hb_local_data);
+
   _mbuffer_head_clear (&session->internals.record_buffer);
   _mbuffer_head_clear (&session->internals.record_recv_buffer);
   _mbuffer_head_clear (&session->internals.record_send_buffer);
@@ -481,12 +489,6 @@ gnutls_deinit (gnutls_session_t session)
 
 /* Returns the minimum prime bits that are acceptable.
  */
-int
-_gnutls_dh_get_allowed_prime_bits (gnutls_session_t session)
-{
-  return session->internals.dh_prime_bits;
-}
-
 int
 _gnutls_dh_set_peer_public (gnutls_session_t session, bigint_t public)
 {
@@ -1306,8 +1308,8 @@ gnutls_handshake_set_post_client_hello_function (gnutls_session_t session,
  *
  * This function can be used to disable certain (security) features in
  * TLS in order to maintain maximum compatibility with buggy
- * clients. It is equivalent to calling:
- * gnutls_record_disable_padding()
+ * clients. Because several trade-offs with security are enabled,
+ * if required they will be reported through the audit subsystem.
  *
  * Normally only servers that require maximum compatibility with
  * everything out there, need to call this function.
@@ -1315,7 +1317,7 @@ gnutls_handshake_set_post_client_hello_function (gnutls_session_t session,
 void
 gnutls_session_enable_compatibility_mode (gnutls_session_t session)
 {
-  gnutls_record_disable_padding (session);
+  ENABLE_COMPAT(&session->internals.priorities);
 }
 
 /**
